import uuid
from typing import List, Dict, Union

import numpy as np

class DTO:
    """
    Implement basic method for DTO objects
    """
    def __hash__(self):
        return hash(tuple(sorted(self.__dict__.items())))

    def __eq__(self, other):
        return isinstance(other, type(self)) and self.__dict__ == other.__dict__

    def __str__(self):
        return "{}({})".format(type(self).__name__, ", ".join(["{}={}".format(k, str(self.__dict__[k])) for k in sorted(self.__dict__)]))

    def __repr__(self):
        return self.__str__()


class Consumption(DTO):
    """
    Consumption element.
    """

    def __init__(self, quantity: Union[np.ndarray, list], cost: int = 0, type: str = ''):
        """
        Create consumption.

        :param quantity: quantity to match
        :param cost: cost of unavailability
        :param type: type of consumption (unique for each node)
        """
        self.cost = cost
        self.quantity = np.array(quantity)
        self.type = type


class Production(DTO):
    """
    Production element
    """
    def __init__(self, quantity: Union[np.ndarray, list], cost: int = 0, type: str = 'in'):
        """
        Create production

        :param quantity: capacity production
        :param cost: cost of use
        :param type: type of production (unique for each node)
        """
        self.type = type
        self.cost = cost
        self.quantity = np.array(quantity)


class Border(DTO):
    """
    Border element
    """
    def __init__(self, dest: str, quantity: Union[np.ndarray, list], cost: int = 0):
        """
        Create border.

        :param dest: node name destination (to export)
        :param quantity: transfer capacity
        :param cost: cost of use
        """
        self.dest = dest
        self.quantity = np.array(quantity)
        self.cost = cost


class InputNode(DTO):
    """
    Node element
    """
    def __init__(self, consumptions: List[Consumption], productions: List[Production], borders: List[Border]):
        """
        Create node element.

        :param consumptions: list of consumptions inside node
        :param productions: list of productions inside node
        :param borders: list of borders inside node
        """
        self.consumptions = consumptions
        self.productions = productions
        self.borders = borders


class Study(DTO):
    """
    Main object to facilitate to build a study
    """

    def __init__(self, node_names=List[str]):
        """
        Instance study.

        :param node_names: list of node names inside network.
        """
        if len(node_names) > len(set(node_names)):
            raise ValueError('some nodes are not unique')

        self._nodes = {name: InputNode(consumptions=[], productions=[], borders=[]) for name in node_names}
        self.horizon = 0


    @property
    def nodes(self):
        return self._nodes

    def add_on_node(self, node: str, data=Union[Production, Consumption, Border]):
        """
        Attach a production or consumption into a node.

        :param node: node name to attach
        :param data: consumption or production to attach
        :return:
        """
        if node not in self._nodes.keys():
            raise ValueError('Node "{}" is not in available nodes'.format(node))

        if isinstance(data, Production):
            self._add_production(node, data)

        elif isinstance(data, Consumption):
            self._add_consumption(node, data)

        return self

    def add_border(self, src: str, dest: str, cost: int, quantity: Union[np.ndarray, List[int]]):
        """
        Add a border inside network.

        :param src: source node name
        :param dest: destination node name
        :param cost: cost of use
        :param quantity: transfer capacity
        :return:
        """
        quantity = np.array(quantity)
        if cost < 0:
            raise ValueError('border cost must be positive')
        if np.all(quantity < 0):
            raise ValueError('border quantity must be positive')
        if dest not in self._nodes.keys():
            raise ValueError('border destination must be a valid node')
        if dest in [b.dest for b in self._nodes[src].borders]:
            raise ValueError('border destination must be unique on a node')
        self._nodes[src].borders.append(Border(dest=dest, quantity=quantity, cost=cost))
        self._update_horizon(quantity)

        return self

    def _add_production(self, node: str, prod: Production):
        if prod.cost < 0:
            raise ValueError('production cost must be positive')
        prod.quantity = np.array(prod.quantity)
        if np.all(prod.quantity < 0):
            raise ValueError('production quantity must be positive')
        if prod.type in [p.type for p in self._nodes[node].productions]:
            raise ValueError('production type must be unique on a node')
        self._nodes[node].productions.append(prod)
        self._update_horizon(prod.quantity)

    def _add_consumption(self, node: str, cons: Consumption):
        if cons.cost < 0:
            raise ValueError('consumption cost must be positive')
        cons.quantity = np.array(cons.quantity)
        if np.all(cons.quantity < 0):
            raise ValueError('consumption quantity must be positive')
        if cons.type in [c.type for c in self._nodes[node].consumptions]:
            raise ValueError('consumption type must be unique on a node')
        self._nodes[node].consumptions.append(cons)
        self._update_horizon(cons.quantity)

    def _update_horizon(self, quantity):
        self.horizon = max(self.horizon, quantity.size)


import unittest
from unittest.mock import MagicMock, call

from hadar.solver.actor.handler.handler import State
from hadar.solver.actor.domain.message import *
from hadar.solver.actor.handler.entry import CanceledCustomerExchangeHandler, ProposalOfferHandler, ProposalHandler
from hadar.solver.actor.ledger import *
from hadar.solver.actor.handler.handler import *


class TestCanceledCustomerExchangeHandler(unittest.TestCase):
    def test_execute_producer(self):
        # Create mock
        tell_mock = MagicMock()
        params = HandlerParameter(tell=tell_mock)

        uuid_mock = MockUUID()

        # Input
        ex_cancel = Exchange(id=10, production_type='nuclear', quantity=10, path_node=['fr', 'be', 'de'])
        ex_keep = Exchange(id=5, production_type='nuclear', quantity=10, path_node=['fr', 'be', 'de'])

        borders = LedgerBorder()
        borders.add(dest='be', cost=2, quantity=10)

        productions = LedgerProduction(uuid_generate=uuid_mock.generate)
        productions.add_production(cost=10, quantity=20, type='nuclear', used=False)

        state = State(name='fr', consumptions=LedgerConsumption(),
                      borders=borders, productions=productions, rac=0, cost=0)
        state.exchanges = LedgerExchange()
        state.exchanges.add_all([ex_cancel, ex_keep], 'export')

        message = ConsumerCanceledExchange(path_node=['fr'], exchanges=[ex_cancel])

        # Expected
        state_exp = State(name='fr', consumptions=LedgerConsumption(),
                          borders=borders, productions=productions, rac=0, cost=0)
        state_exp.exchanges = LedgerExchange()
        state_exp.exchanges.add(ex_keep, 'export')

        # Test
        handler = CanceledCustomerExchangeHandler(params=params)
        res, _ = handler.execute(state=state, message=message)

        self.assertEqual(state_exp, res)
        tell_mock.assert_called_with(to='be',
                                     mes=Proposal(production_type='nuclear', cost=12, quantity=10, path_node=['fr']))


class TestProposalOfferHandler(unittest.TestCase):
    def test_execute_produce(self):
        # Create mock
        ask_mock = MagicMock()
        uuid_mock = MockUUID()
        params = HandlerParameter(ask=ask_mock, uuid_generate=uuid_mock.generate)

        # Input
        productions = LedgerProduction(uuid_generate=uuid_mock.generate)
        productions.add_production(cost=10, quantity=10, type='solar')

        borders = LedgerBorder()
        borders.add(dest='be', cost=10, quantity=8)

        state = State(name='fr', consumptions=None, borders=borders, productions=productions, rac=0, cost=0)
        state.exchanges = LedgerExchange()
        state.exchanges.add_all([Exchange(id=1, production_type='solar', quantity=1, path_node=['be']),
                                 Exchange(id=2, production_type='solar', quantity=1, path_node=['be']),
                                 Exchange(id=3, production_type='solar', quantity=1, path_node=['be']),
                                 Exchange(id=4, production_type='solar', quantity=1, path_node=['be']),
                                 Exchange(id=5, production_type='solar', quantity=1, path_node=['be'])], 'export')

        offer = ProposalOffer(production_type='solar', cost=10, quantity=5, path_node=['fr'],
                              return_path_node=['fr', 'be'])

        # Expected
        message_exp = [Exchange(id=6, production_type='solar', quantity=1, path_node=['fr', 'be']),
                       Exchange(id=7, production_type='solar', quantity=1, path_node=['fr', 'be']),
                       Exchange(id=8, production_type='solar', quantity=1, path_node=['fr', 'be'])]
        exchanges_exp = [Exchange(id=6, production_type='solar', quantity=1, path_node=['be']),
                         Exchange(id=7, production_type='solar', quantity=1, path_node=['be']),
                         Exchange(id=8, production_type='solar', quantity=1, path_node=['be'])]
        state_exp = deepcopy(state)
        state_exp.exchanges.add_all(exchanges_exp, type='export')

        # Test
        handler = ProposalOfferHandler(params=params)
        state_res, message_res = handler.execute(state=state, message=offer)

        self.assertEqual(state_exp, state_res)
        self.assertEqual(message_exp, message_res)

    def test_execute_backward(self):
        # Create mock
        message_mock = [Exchange(id=2, production_type='solar', quantity=1, path_node=['de', 'fr', 'be']),
                        Exchange(id=3, production_type='solar', quantity=1, path_node=['de', 'fr', 'be']),
                        Exchange(id=4, production_type='solar', quantity=1, path_node=['de', 'fr', 'be'])]
        ask_mock = MagicMock(return_value=message_mock)
        uuid_mock = MockUUID()
        params = HandlerParameter(ask=ask_mock, uuid_generate=uuid_mock.generate)

        # Input
        productions = LedgerProduction(uuid_generate=uuid_mock.generate)
        productions.add_production(cost=10, quantity=10, type='solar')

        borders = LedgerBorder()
        borders.add(dest='be', cost=10, quantity=8)

        state = State(name='fr', consumptions=None, borders=borders, productions=productions, rac=0, cost=0)
        state.exchanges = LedgerExchange()
        state.exchanges.add(Exchange(id=0, production_type='solar', quantity=5, path_node=['be']), type='transfer')

        offer = ProposalOffer(production_type='solar', cost=10, quantity=5, path_node=['fr', 'de'],
                              return_path_node=['de', 'fr', 'be'])

        # Expected
        exchanges_exp = [Exchange(id=2, production_type='solar', quantity=1, path_node=['be']),
                         Exchange(id=3, production_type='solar', quantity=1, path_node=['be']),
                         Exchange(id=4, production_type='solar', quantity=1, path_node=['be'])]
        state_exp = deepcopy(state)
        state_exp.exchanges.add_all(exchanges_exp, type='transfer')

        # Test
        handler = ProposalOfferHandler(params=params)
        state_res, message_res = handler.execute(state=state, message=offer)

        self.assertEqual(state_exp, state_res)
        self.assertEqual(message_mock, message_res)


class TestProposalHandler(unittest.TestCase):
    def test_execute(self):
        # Mock
        expected_exs = [Exchange(id=100, quantity=1, production_type='solar', path_node=['it']),
                        Exchange(id=101, quantity=1, production_type='solar', path_node=['it']),
                        Exchange(id=102, quantity=1, production_type='solar', path_node=['it']),
                        Exchange(id=103, quantity=1, production_type='solar', path_node=['it']),
                        Exchange(id=104, quantity=1, production_type='solar', path_node=['it'])]
        ask_mock = MagicMock(return_value=expected_exs)
        tell_mock = MagicMock()
        params = HandlerParameter(ask=ask_mock, tell=tell_mock)

        uuid_mock = MockUUID()

        # Input
        consumptions = LedgerConsumption()
        consumptions.add(type='load', quantity=10, cost=10 ** 6)

        productions = LedgerProduction(uuid_generate=uuid_mock.generate)
        productions.add_production(type='solar', cost=5, quantity=5, used=True)
        productions.add_production(type='nuclear', cost=10, quantity=5, used=True)

        borders = LedgerBorder()
        borders.add(dest='be', cost=2, quantity=10)

        state = State(name='fr', consumptions=consumptions, borders=borders, productions=productions, rac=0, cost=75)
        proposal = Proposal(production_type='wind', cost=7, quantity=10, path_node=['it'])

        # Expected
        proposal_forward = Proposal(production_type='wind', cost=9, quantity=5, path_node=['fr', 'it'])
        proposal_nuclear = Proposal(production_type='nuclear', cost=12, quantity=5, path_node=['fr'])
        offer_expected = ProposalOffer(production_type='wind', cost=7, quantity=5, path_node=['it'],
                                       return_path_node=['fr'])

        state_expected = deepcopy(state)
        state_expected.cost = 60
        state_expected.exchanges.add_all(expected_exs, type='import')
        state_expected.productions.ledger = \
            pd.DataFrame({'cost': [5] * 5 + [7] * 5 + [10] * 5,
                          'quantity': [1] * 15,
                          'type': ['solar'] * 5 + ['import'] * 5 + ['nuclear'] * 5,
                          'used': [True] * 10 + [False] * 5,
                          'path_node': [None, None, None, None, None,
                                        ['it'], ['it'], ['it'], ['it'], ['it'],
                                        None, None, None, None, None]},
                         index=[1, 2, 3, 4, 5, 100, 101, 102, 103, 104, 6, 7, 8, 9, 10])

        # Test
        handler = ProposalHandler(params=params)
        state_res, _ = handler.execute(deepcopy(state), deepcopy(proposal))
        self.assertEqual(state_expected, state_res)

        ask_mock.assert_called_with(to='it', mes=offer_expected)
        tell_mock.assert_has_calls(
            [call(to='be', mes=proposal_forward),
             call(to='be', mes=proposal_nuclear)])


class MockUUID:
    def __init__(self):
        self.inc = 0

    def generate(self):
        self.inc += 1
        return self.inc



from abc import ABC, abstractmethod


class ABCPlotting(ABC):
    """
    Abstract method to plot solver result.
    """

    @abstractmethod
    def stack(self, node: str):
        pass

    @abstractmethod
    def exchanges_map(self, t: int, limit: int):
        pass


import numpy as np

from typing import Union, List, Dict

from hadar.solver.input import *


class DTO:
    """
    Implement basic method for DTO objects
    """

    def __hash__(self):
        return hash(tuple(sorted(self.__dict__.items())))

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        for name, att in self.__dict__.items():
            if isinstance(att, (np.ndarray, np.generic)):
                if not np.array_equal(self.__dict__[name], other.__dict__[name]):
                    return False
            elif self.__dict__[name] != other.__dict__[name]:
                return False
        return True

    def __str__(self):
        return "{}({})".format(type(self).__name__,
                               ", ".join(["{}={}".format(k, str(self.__dict__[k])) for k in sorted(self.__dict__)]))

    def __repr__(self):
        return self.__str__()


class OutputConsumption(DTO):
    """
    Consumption element
    """
    def __init__(self, quantity: Union[np.ndarray, list], cost: int = 0, type: str = ''):
        """
        Create instance.

        :param quantity: quantity matched by node
        :param cost: cost of unavailability
        :param type: consumption type (unique in a node)
        """
        self.cost = cost
        self.quantity = np.array(quantity)
        self.type = type


class OutputProduction(DTO):
    """
    Production element
    """
    def __init__(self, quantity: Union[np.ndarray, list], cost: int = 0, type: str = 'in'):
        """
        Create instance.

        :param quantity: capacity used by node
        :param cost: cost of use
        :param type: production type (unique in a node)
        """
        self.type = type
        self.cost = cost
        self.quantity = np.array(quantity)


class OutputBorder(DTO):
    """
    Border element
    """
    def __init__(self, dest: str, quantity: Union[np.ndarray, list], cost: int = 0):
        """
        Create instance.

        :param dest: destination node name
        :param quantity: capacity used
        :param cost: cost of use
        """
        self.dest = dest
        self.quantity = np.array(quantity)
        self.cost = cost


class OutputNode(DTO):
    """
    Node element
    """
    def __init__(self,
                 consumptions: List[OutputConsumption],
                 productions: List[OutputProduction],
                 borders: List[OutputBorder]):
        """
        Create Node.

        :param consumptions: consumptions list
        :param productions: productions list
        :param borders:  border list
        """
        self.consumptions = consumptions
        self.productions = productions
        self.borders = borders

    @staticmethod
    def build_like_input(input: InputNode):
        """
        Use an input node to create an output node. Keep list elements fill quantity by zeros.

        :param input: InputNode to copy
        :return: OutputNode like InputNode with all quantity at zero
        """
        output = OutputNode(consumptions=[], productions=[], borders=[])

        output.consumptions = [OutputConsumption(type=i.type, cost=i.cost, quantity=np.zeros_like(i.quantity))
                               for i in input.consumptions]
        output.productions = [OutputProduction(type=i.type, cost=i.cost, quantity=np.zeros_like(i.quantity))
                              for i in input.productions]
        output.borders = [OutputBorder(dest=i.dest, cost=i.cost, quantity=np.zeros_like(i.quantity))
                          for i in input.borders]
        return output


class Result(DTO):
    """
    Result of study
    """
    def __init__(self, nodes: Dict[str, OutputNode]):
        """
        Create result
        :param nodes: list of nodes present in network
        """
        self._nodes = nodes

    @property
    def nodes(self):
        return self._nodes
